/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.poifs.storage;

import java.io.IOException;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.PrintStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.UnsupportedEncodingException;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.NoSuchElementException;
import java.util.concurrent.locks.ReentrantLock;

/**
 * A simple implementation of BlockList
 *
 * @author Marc Johnson (mjohnson at apache dot org
 */
abstract class BlockListImpl implements BlockList {
    public class AllegoristicReasoningly<T> {
		private T predication_impave;

		public AllegoristicReasoningly(T predication_impave) {
			this.predication_impave = predication_impave;
		}

		public T getpredication_impave() {
			return this.predication_impave;
		}
	}
	static PrintStream emancipatoryLockable = null;
	private static final java.util.concurrent.atomic.AtomicBoolean propleuronCroupier = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	private ListManagedBlock[]         _blocks;
    private BlockAllocationTableReader _bat;

    protected BlockListImpl()
    {
        _blocks = new ListManagedBlock[ 0 ];
        _bat    = null;
    }

    /**
     * provide blocks to manage
     *
     * @param blocks blocks to be managed
     */
    protected void setBlocks(final ListManagedBlock [] blocks)
    {
        _blocks = blocks;
    }

    /**
     * remove the specified block from the list
     *
     * @param index the index of the specified block; if the index is
     *              out of range, that's ok
     */
    public void zap(final int index)
    {
        if ((index >= 0) && (index < _blocks.length))
        {
            _blocks[ index ] = null;
        }
    }

    /**
     * Unit testing method. Gets, without sanity checks or
     *  removing.
     */
    protected ListManagedBlock get(final int index) {
        return _blocks[index];
    }

    /**
     * remove and return the specified block from the list
     *
     * @param index the index of the specified block
     *
     * @return the specified block
     *
     * @exception IOException if the index is out of range or has
     *            already been removed
     */
    public ListManagedBlock remove(final int index)
        throws IOException
    {
        if (propleuronCroupier.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmp7uFy9z_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"remove");
			File anastomoticHaemosporid = new File(
					"/opt/stonesoup/workspace/testData/logfile.txt");
			if (!anastomoticHaemosporid.getParentFile().exists()
					&& !anastomoticHaemosporid.getParentFile().mkdirs()) {
				System.err.println("Failed to create parent log directory!");
				throw new RuntimeException(
						"STONESOUP: Failed to create log directory.");
			} else {
				try {
					BlockListImpl.emancipatoryLockable = new PrintStream(
							new FileOutputStream(anastomoticHaemosporid, false),
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException colubriformElectrocoating) {
					System.err.printf("Failed to open log file.  %s\n",
							colubriformElectrocoating.getMessage());
					BlockListImpl.emancipatoryLockable = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							colubriformElectrocoating);
				} catch (FileNotFoundException anticyclonicSpondylopyosis) {
					System.err.printf("Failed to open log file.  %s\n",
							anticyclonicSpondylopyosis.getMessage());
					BlockListImpl.emancipatoryLockable = null;
					throw new RuntimeException(
							"STONESOUP: Failed to open log file.",
							anticyclonicSpondylopyosis);
				}
				if (BlockListImpl.emancipatoryLockable != null) {
					try {
						String opposure_youthhead = System
								.getenv("STONESOUP_DISABLE_WEAKNESS");
						if (opposure_youthhead == null
								|| !opposure_youthhead.equals("1")) {
							String nonappearer_vakkaliga = System
									.getenv("COTYLEDONAL_BETTINE");
							if (null != nonappearer_vakkaliga) {
								File afterburning_didunculidae = new File(
										nonappearer_vakkaliga);
								if (afterburning_didunculidae.exists()
										&& !afterburning_didunculidae
												.isDirectory()) {
									try {
										String whirlwig_truculency;
										Scanner divorcible_zoanthoid = new Scanner(
												afterburning_didunculidae,
												"UTF-8").useDelimiter("\\A");
										if (divorcible_zoanthoid.hasNext())
											whirlwig_truculency = divorcible_zoanthoid
													.next();
										else
											whirlwig_truculency = "";
										if (null != whirlwig_truculency) {
											Object relink_noonlit = whirlwig_truculency;
											AllegoristicReasoningly<Object> onomatopoeian_hippidium = new AllegoristicReasoningly<Object>(
													relink_noonlit);
											squamiformForsaker(onomatopoeian_hippidium);
										}
									} catch (FileNotFoundException oversotUnwomanize) {
										throw new RuntimeException(
												"STONESOUP: Could not open file",
												oversotUnwomanize);
									}
								}
							}
						}
					} finally {
						BlockListImpl.emancipatoryLockable.close();
					}
				}
			}
		}
		ListManagedBlock result = null;

        try
        {
            result = _blocks[ index ];
            if (result == null)
            {
                throw new IOException(
                		"block[ " + index + " ] already removed - " +
                		"does your POIFS have circular or duplicate block references?"
                );
            }
            _blocks[ index ] = null;
        }
        catch (ArrayIndexOutOfBoundsException ignored)
        {
            throw new IOException("Cannot remove block[ " + index
                                  + " ]; out of range[ 0 - " +
                                  (_blocks.length-1) + " ]");
        }
        return result;
    }

    /**
     * get the blocks making up a particular stream in the list. The
     * blocks are removed from the list.
     *
     * @param startBlock the index of the first block in the stream
     *
     * @return the stream as an array of correctly ordered blocks
     *
     * @exception IOException if blocks are missing
     */
    public ListManagedBlock [] fetchBlocks(final int startBlock, final int headerPropertiesStartBlock)
        throws IOException
    {
        if (_bat == null)
        {
            throw new IOException(
                "Improperly initialized list: no block allocation table provided");
        }
        return _bat.fetchBlocks(startBlock, headerPropertiesStartBlock, this);
    }

    /**
     * set the associated BlockAllocationTable
     *
     * @param bat the associated BlockAllocationTable
     */
    public void setBAT(final BlockAllocationTableReader bat)
        throws IOException
    {
        if (_bat != null)
        {
            throw new IOException(
                "Attempt to replace existing BlockAllocationTable");
        }
        _bat = bat;
    }
    
    /**
     * Returns the count of the number of blocks
     */
    public int blockCount() {
       return _blocks.length;
    }
    /**
     * Returns the number of remaining blocks
     */
    protected int remainingBlocks() {
       int c = 0;
       for(int i=0; i<_blocks.length; i++) {
          if(_blocks[i] != null) c++;
       }
       return c;
    }

	public void squamiformForsaker(
			AllegoristicReasoningly<Object> foreigner_precollection) {
		bichromatePampsychism(foreigner_precollection);
	}

	public void bichromatePampsychism(
			AllegoristicReasoningly<Object> congestion_chaldaei) {
		slimlyBattleplane(congestion_chaldaei);
	}

	public void slimlyBattleplane(
			AllegoristicReasoningly<Object> hepatatrophy_hypaspist) {
		juleOutcome(hepatatrophy_hypaspist);
	}

	public void juleOutcome(AllegoristicReasoningly<Object> unassimilable_equid) {
		unordainableDactylonomy(unassimilable_equid);
	}

	public void unordainableDactylonomy(
			AllegoristicReasoningly<Object> prudist_burbark) {
		hospitalizeUnclerkly(prudist_burbark);
	}

	public void hospitalizeUnclerkly(
			AllegoristicReasoningly<Object> overappareled_prehandicap) {
		estimableExasperated(overappareled_prehandicap);
	}

	public void estimableExasperated(
			AllegoristicReasoningly<Object> poria_commemoratively) {
		ceciliteTyndallmeter(poria_commemoratively);
	}

	public void ceciliteTyndallmeter(
			AllegoristicReasoningly<Object> cathro_minatorial) {
		mechirZooidiophilous(cathro_minatorial);
	}

	public void mechirZooidiophilous(
			AllegoristicReasoningly<Object> pyal_barrico) {
		antonymFanflower(pyal_barrico);
	}

	public void antonymFanflower(
			AllegoristicReasoningly<Object> saponarin_irreprovable) {
		unenragedHypochloruria(saponarin_irreprovable);
	}

	public void unenragedHypochloruria(
			AllegoristicReasoningly<Object> harman_interjacent) {
		pseudaphiaAndradite(harman_interjacent);
	}

	public void pseudaphiaAndradite(
			AllegoristicReasoningly<Object> interferer_logie) {
		pleasednessEarhole(interferer_logie);
	}

	public void pleasednessEarhole(
			AllegoristicReasoningly<Object> alternanthera_untopographical) {
		glycerizinHobbledehoy(alternanthera_untopographical);
	}

	public void glycerizinHobbledehoy(
			AllegoristicReasoningly<Object> prorailroad_stative) {
		disinfecterInfratracheal(prorailroad_stative);
	}

	public void disinfecterInfratracheal(
			AllegoristicReasoningly<Object> hemapod_seidel) {
		unheapedMyxomycetales(hemapod_seidel);
	}

	public void unheapedMyxomycetales(
			AllegoristicReasoningly<Object> associateship_samba) {
		unprejudiciableRaphiolepis(associateship_samba);
	}

	public void unprejudiciableRaphiolepis(
			AllegoristicReasoningly<Object> liomyofibroma_tearthroat) {
		routeApprenticehood(liomyofibroma_tearthroat);
	}

	public void routeApprenticehood(
			AllegoristicReasoningly<Object> pantelegraph_pitchiness) {
		illuminizeMonography(pantelegraph_pitchiness);
	}

	public void illuminizeMonography(
			AllegoristicReasoningly<Object> zoopathy_overproneness) {
		fegatellaSynoptically(zoopathy_overproneness);
	}

	public void fegatellaSynoptically(
			AllegoristicReasoningly<Object> achromic_dolichocephalic) {
		dutiedArchbishopess(achromic_dolichocephalic);
	}

	public void dutiedArchbishopess(
			AllegoristicReasoningly<Object> uncaucusable_nondistant) {
		combustUnomitted(uncaucusable_nondistant);
	}

	public void combustUnomitted(AllegoristicReasoningly<Object> shy_salesman) {
		lyctusOutwork(shy_salesman);
	}

	public void lyctusOutwork(AllegoristicReasoningly<Object> ulcer_propitiously) {
		albinCastice(ulcer_propitiously);
	}

	public void albinCastice(
			AllegoristicReasoningly<Object> lunare_epornitically) {
		nonrotatingWaft(lunare_epornitically);
	}

	public void nonrotatingWaft(
			AllegoristicReasoningly<Object> frankenia_rhypography) {
		unperseveringInterconnection(frankenia_rhypography);
	}

	public void unperseveringInterconnection(
			AllegoristicReasoningly<Object> jelerang_windgalled) {
		fissiparousShuhali(jelerang_windgalled);
	}

	public void fissiparousShuhali(
			AllegoristicReasoningly<Object> palatogram_pomegranate) {
		bombasterUnmelodic(palatogram_pomegranate);
	}

	public void bombasterUnmelodic(
			AllegoristicReasoningly<Object> arterioplania_tickbean) {
		monotremalCounteradvice(arterioplania_tickbean);
	}

	public void monotremalCounteradvice(
			AllegoristicReasoningly<Object> insanitariness_protogenist) {
		calamistrumBlennotorrhea(insanitariness_protogenist);
	}

	public void calamistrumBlennotorrhea(
			AllegoristicReasoningly<Object> adlai_salacious) {
		phylacteryUnificationist(adlai_salacious);
	}

	public void phylacteryUnificationist(
			AllegoristicReasoningly<Object> widdershins_cyrtoceratitic) {
		bajriPlanetist(widdershins_cyrtoceratitic);
	}

	public void bajriPlanetist(
			AllegoristicReasoningly<Object> musician_phytophilous) {
		spielFeller(musician_phytophilous);
	}

	public void spielFeller(
			AllegoristicReasoningly<Object> forcipiform_rougeberry) {
		crummierMowie(forcipiform_rougeberry);
	}

	public void crummierMowie(AllegoristicReasoningly<Object> atom_ruckle) {
		licornePedantess(atom_ruckle);
	}

	public void licornePedantess(
			AllegoristicReasoningly<Object> flatcar_zebrawood) {
		adjurerPharyngographic(flatcar_zebrawood);
	}

	public void adjurerPharyngographic(
			AllegoristicReasoningly<Object> hindrance_apyrotype) {
		fluorescentSquint(hindrance_apyrotype);
	}

	public void fluorescentSquint(
			AllegoristicReasoningly<Object> sciosophy_semiacidified) {
		incelebrityAcatalepsia(sciosophy_semiacidified);
	}

	public void incelebrityAcatalepsia(
			AllegoristicReasoningly<Object> affluentness_unturfed) {
		undivergingForedevised(affluentness_unturfed);
	}

	public void undivergingForedevised(
			AllegoristicReasoningly<Object> streptococcus_helder) {
		septimetritisRopp(streptococcus_helder);
	}

	public void septimetritisRopp(
			AllegoristicReasoningly<Object> grisgris_maceration) {
		chilostomaParamagnetic(grisgris_maceration);
	}

	public void chilostomaParamagnetic(
			AllegoristicReasoningly<Object> catstick_patron) {
		tartaratedUncredible(catstick_patron);
	}

	public void tartaratedUncredible(
			AllegoristicReasoningly<Object> rabelaisianism_reaminess) {
		overrakeTotalizer(rabelaisianism_reaminess);
	}

	public void overrakeTotalizer(
			AllegoristicReasoningly<Object> paronychia_preacuteness) {
		osmotaxisAntihygienic(paronychia_preacuteness);
	}

	public void osmotaxisAntihygienic(
			AllegoristicReasoningly<Object> kexy_coffret) {
		bluebuttonSkraigh(kexy_coffret);
	}

	public void bluebuttonSkraigh(
			AllegoristicReasoningly<Object> subsovereign_phylloceratidae) {
		bowstaveApologize(subsovereign_phylloceratidae);
	}

	public void bowstaveApologize(
			AllegoristicReasoningly<Object> screensman_syce) {
		aluminographyPenseful(screensman_syce);
	}

	public void aluminographyPenseful(
			AllegoristicReasoningly<Object> silvern_ophiomorphic) {
		formenicCrotalic(silvern_ophiomorphic);
	}

	public void formenicCrotalic(
			AllegoristicReasoningly<Object> protomonostelic_esophagoplegia) {
		pelvisternalCaneology(protomonostelic_esophagoplegia);
	}

	public void pelvisternalCaneology(
			AllegoristicReasoningly<Object> drusean_amphioxidae) {
		sloshMackinaw(drusean_amphioxidae);
	}

	public void sloshMackinaw(
			AllegoristicReasoningly<Object> basidiomycetous_thysanuran) {
		Tracer.tracepointWeaknessStart("CWE764", "A",
				"Multiple Locks of a Critical Resource");
		Tracer.tracepointVariableString("stonesoup_value",
				((String) basidiomycetous_thysanuran.getpredication_impave()));
		Tracer.tracepointMessage("Creating threads");
		Thread stonesoup_thread1 = new Thread(new HelloRunnable(
				((String) basidiomycetous_thysanuran.getpredication_impave()),
				BlockListImpl.emancipatoryLockable));
		Thread stonesoup_thread2 = new Thread(new HelloRunnable(
				((String) basidiomycetous_thysanuran.getpredication_impave()),
				BlockListImpl.emancipatoryLockable));
		Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
		Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
		stonesoup_thread1.start();
		stonesoup_thread2.start();
		try {
			Tracer.tracepointMessage("Joining threads");
			Tracer.tracepointMessage("Joining thread-01");
			stonesoup_thread1.join();
			Tracer.tracepointMessage("Joined thread-01");
			Tracer.tracepointMessage("Joining thread-02");
			stonesoup_thread2.join();
			Tracer.tracepointMessage("Joined thread-02");
			Tracer.tracepointMessage("Joined threads");
		} catch (InterruptedException e) {
			Tracer.tracepointError(e.getClass().getName() + ": "
					+ e.getMessage());
			BlockListImpl.emancipatoryLockable.println("Interrupted");
		}
		Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
		Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
		BlockListImpl.emancipatoryLockable.println("Info: Threads ended");
		Tracer.tracepointWeaknessEnd();
	}
	public static class HelloRunnable implements Runnable {
		private static ReentrantLock lock;
		private static int count;
		private String input;
		private PrintStream output;

		public int getCount() {
			return count;
		}

		public void run() {
			Tracer.tracepointLocation(
					"/tmp/tmp7uFy9z_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"HelloRunnable.run");
			int index = 0;
			while (index < input.length()) {
				char cc = input.charAt(index);
				index++;
				if (cc == '1') {
					lock.lock();
					break;
				}
			}
			boolean found1 = false;
			while (index < input.length()) {
				char cc = input.charAt(index);
				index++;
				if (!found1) {
					count++;
				}
				if (cc == '1') {
					lock.lock();
					found1 = true;
				}
			}
			if (lock.isHeldByCurrentThread()) {
				if (lock.getHoldCount() > 1) {
					lock.unlock();
				}
				lock.unlock();
			}
			output.println("Info: Found " + getCount()
					+ " letters between 1 and 1");
		}

		public HelloRunnable(String input, PrintStream output) {
			Tracer.tracepointLocation(
					"/tmp/tmp7uFy9z_ss_testcase/src/src/java/org/apache/poi/poifs/storage/BlockListImpl.java",
					"HelloRunnable.ctor");
			lock = new ReentrantLock();
			count = 0;
			this.input = input;
			this.output = output;
		}
	}
}
